unit Graph3;

interface

uses
  Windows, SysUtils, Classes, Controls, Graphics, Contnrs, Math, Forms;

type
  TDoubleArray = array of Double;

  //TObject is de "domme" basisklasse waar alles vanaf stamt. TObject kan en
  //doet niets en is prima geschikt om slechts een "formuletje" in op te slaan
  TEquation = class(TObject)
  private
    FA: Double;
    FB: Double;
    FC: Double;
    FColor: TColor;
    FD: Double;
    FName: String;
    FOnChanged: TNotifyEvent;
    FVisible: Boolean;
    procedure Changed;
    procedure SetA(Value: Double);
    procedure SetB(Value: Double);
    procedure SetC(Value: Double);
    procedure SetColor(Value: TColor);
    procedure SetD(Value: Double);
    procedure SetVisible(Value: Boolean);
  public
    constructor Create;
    function Derivative1(IncludeName: Boolean = False): String;
    function Formula(IncludeName: Boolean = False): String;
    procedure GetTangent(X, XA, XB: Double; var YA, YB: Double);
    procedure GetZeroPoints(XA, XB: Double; var XCoords: TDoubleArray);
    function HasTangents: Boolean;
    function Y(X: Double): Double;
    function Y1(X: Double): Double;
    property A: Double read FA write SetA;
    property B: Double read FB write SetB;
    property C: Double read FC write SetC;
    property Color: TColor read FColor write SetColor;
    property D: Double read FD write SetD;
    property Name: String read FName write FName;
    property OnChanged: TNotifyEvent read FOnChanged write FOnChanged;
    property Visible: Boolean read FVisible write SetVisible;
  end;

  //Een eigen enumeratie-type voor de tekenopties
  TPaintOption = (poAxes, poScale, poFormula, poDerivative1, poZeroPoints);
  TPaintOptions = set of TPaintOption;

  //TGraphicControl is een visueel basiscontrol met een Canvas
  TGraph = class(TGraphicControl)
  private
    FAxisColor: TColor;
    FEquations: TList;
    FPaintOptions: TPaintOptions;
    //De omrekenschaal van formule naar pixels, gebaseerd op de breedte
    FScale: Single;
    //De real lengte van de zichtbare x-as
    FXRange: Double;
    //Gekoppeld aan de OnChanged property van elke TEquation in FEquations
    procedure EquationChanged(Sender: TObject);
    function GetEquation(Index: Integer): TEquation;
    function GetEquationCount: Integer;
    //De "setters" voor de property's AxisColor, PaintOptions, Scale en XRange
    procedure SetAxisColor(Value: TColor);
    procedure SetPaintOptions(Value: TPaintOptions);
    procedure SetScale(Value: Single);
    procedure SetXRange(Value: Double);
  protected
    //Paint wordt automatisch door de VCL aangeroepen wanneer dit control
    //zichzelf moet hertekenen, maar is ook te triggeren d.m.v. bijv. Invalidate
    procedure Paint; override;
    //Resize wordt automatisch door de VCL aangeroepen wanner dit control van
    //formaat is gewijzigd
    procedure Resize; override;
  public
    procedure AddEquation(A, B, C, D: Double; AColor: TColor = clDefault;
      AName: String = 'f');
    procedure AnimateTangents(EquationIndex: Integer);
    procedure Clear;
    //In de constructor initialiseer je diverse instellingen en maak je de
    //eigen interne objecten aan...
    constructor Create(AOwner: TComponent); override;
    //... en in de destructor ruim je ze weer op
    destructor Destroy; override;
    property AxisColor: TColor read FAxisColor write SetAxisColor;
    property EquationCount: Integer read GetEquationCount;
    property Equations[Index: Integer]: TEquation read GetEquation;
    property PaintOptions: TPaintOptions read FPaintOptions
      write SetPaintOptions;
    property Scale: Single read FScale write SetScale;
    property XRange: Double read FXRange write SetXRange;
  public
    //Deze property's zitten standaard al in TGraphicControl, maar zijn nog
    //verborgen, dus even "publiek" maken
    property Color;
    property Font;
  end;

implementation

{ TEquation }

procedure TEquation.Changed;
begin
  if Assigned(FOnChanged) then
    FOnChanged(Self);
end;

constructor TEquation.Create;
begin
  inherited Create;
  FVisible := True;
end;

function TEquation.Derivative1(IncludeName: Boolean = False): String;
begin
  Result := FloatToStr(3 * A) + 'x'#178;
  if B < 0 then
    Result := Result + FloatToStr(2 * B) + 'x'
  else
    Result := Result + '+' + FloatToStr(2 * B) + 'x';
  if C < 0 then
    Result := Result + FloatToStr(C)
  else
    Result := Result + '+' + FloatToStr(C);
  if IncludeName then
    Result := FName + '''(x) = ' + Result;
end;

function TEquation.Formula(IncludeName: Boolean = False): String;
begin
  Result := FloatToStr(A) + 'x'#179;
  if B < 0 then
    Result := Result + FloatToStr(B) + 'x'#178
  else
    Result := Result + '+' + FloatToStr(B) + 'x'#178;
  if C < 0 then
    Result := Result + FloatToStr(C) + 'x'
  else
    Result := Result + '+' + FloatToStr(C) + 'x';
  if D < 0 then
    Result := Result + FloatToStr(D)
  else
    Result := Result + '+' + FloatToStr(D);
  if IncludeName then
    Result := FName + '(x) = ' + Result;
end;

procedure TEquation.GetTangent(X, XA, XB: Double; var YA, YB: Double);
var
  Y: Double;
  Y1: Double;
begin
  Y := Self.Y(X);
  Y1 := Self.Y1(X);
  YA := Y + Y1 * (XA - X);
  YB := Y + Y1 * (XB - X);
end;

procedure TEquation.GetZeroPoints(XA, XB: Double; var XCoords: TDoubleArray);
var
  X1: Double;
  Y1: Double;
  X2: Double;
  Y2: Double;
  Step: Double;
begin
  if (FA <> 0) or (FB <> 0) or (FC <> 0) then
  begin
    SetLength(XCoords, 0);
    X1 := XA;
    Y1 := Y(X1);
    Step := (XB - XA) / 100;
    while X1 < XB do
    begin
      X2 := X1 + Step;
      Y2 := Y(X2);
      if Sign(Y1) <> Sign(Y2) then
      begin
        Step := (X2 - X1) / 100;
        X2 := X1 + Step;
        Y2 := Y(X2);
        while Abs(Y2) < Abs(Y1) do
        begin
          X1 := X2;
          Y1 := Y2;
          X2 := X1 + Step;
          Y2 := Y(X2);
        end;
        SetLength(XCoords, Length(XCoords) + 1);
        XCoords[Length(XCoords) - 1] := X2;
        Step := (XB - XA) / 100;
      end;
      X1 := X2;
      Y1 := Y2;
    end;
  end;
end;

function TEquation.HasTangents: Boolean;
begin
  Result := (FA <> 0) or (FB <> 0);
end;

procedure TEquation.SetA(Value: Double);
begin
  if FA <> Value then
  begin
    FA := Value;
    Changed;
  end;
end;

procedure TEquation.SetB(Value: Double);
begin
  if FB <> Value then
  begin
    FB := Value;
    Changed;
  end;
end;

procedure TEquation.SetC(Value: Double);
begin
  if FC <> Value then
  begin
    FC := Value;
    Changed;
  end;
end;

procedure TEquation.SetColor(Value: TColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    Changed;
  end;
end;

procedure TEquation.SetD(Value: Double);
begin
  if FD <> Value then
  begin
    FD := Value;
    Changed;
  end;
end;

procedure TEquation.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    Changed;
  end;
end;

function TEquation.Y(X: Double): Double;
begin
  Result := FA * Power(X, 3) + FB * Power(X, 2) + FC * X + FD;
end;

function TEquation.Y1(X: Double): Double;
begin
  Result := 3 * FA * Power(X, 2) + 2 * FB * X + FC;
end;

{ TGraph }

procedure TGraph.AddEquation(A, B, C, D: Double; AColor: TColor = clDefault;
  AName: String = 'f');
var
  Equation: TEquation;
begin
  Equation := TEquation.Create;
  Equation.A := A;
  Equation.B := B;
  Equation.C := C;
  Equation.D := D;
  Equation.Color := AColor;
  Equation.Name := AName;
  Equation.OnChanged := EquationChanged;
  FEquations.Add(Equation);
  Invalidate;
end;

procedure TGraph.AnimateTangents(EquationIndex: Integer);
var
  Equation: TEquation;
  P: TPoint;
  XA: Double;
  XB: Double;
  X: Double;
  YA: Double;
  YB: Double;
  PA: TPoint;
  PB: TPoint;
begin
  Equation := Equations[EquationIndex];
  if (Width > 1) and (Height > 1) and Equation.HasTangents then
  begin
    P.X := 0;
    XA := (-Width div 2) * FScale;
    XB := (Width div 2) * FScale;
    PA.X := 0;
    PB.X := Width;
    while (not Application.Terminated) and (P.X < Width) do
    begin
      Canvas.Pen.Color := Equation.Color;
      X := (P.X - Width div 2) * FScale;
      Equation.GetTangent(X, XA, XB, YA, YB);
      PA.Y := -Round(YA / FScale) + Height div 2;
      PB.Y := -Round(YB / FScale) + Height div 2;
      Canvas.MoveTo(PA.X, PA.Y);
      Canvas.LineTo(PB.X, PB.Y);
      Inc(P.X, 2);
      Sleep(5);
      Application.ProcessMessages;
    end;
  end;
end;

procedure TGraph.Clear;
begin
  FEquations.Clear;
  Invalidate;
end;

constructor TGraph.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  //ControlStyle instellen is (in dit geval) geen vereiste
  ControlStyle := [csOpaque, csNoStdEvents];
  //TList is een object waar je andere objecten in kunt bijhouden. TObjectList
  //stamt af van TList en ruimt de objecten die er in zitten ook weer
  //automatisch op
  FEquations := TObjectList.Create(True);
  FPaintOptions := [poAxes, poScale, poFormula, poZeroPoints];
  Width := 100;
  Height := 100;
  XRange := 10.0;
end;

destructor TGraph.Destroy;
begin
  FEquations.Free;
  inherited Destroy;
end;

procedure TGraph.EquationChanged(Sender: TObject);
begin
  Invalidate;
end;

function TGraph.GetEquation(Index: Integer): TEquation;
begin
  Result := FEquations[Index];
end;

function TGraph.GetEquationCount: Integer;
begin
  Result := FEquations.Count;
end;

procedure TGraph.Paint;
type
  TFunc = function(X: Double): Double of object;
var
  I: Integer;
  Equation: TEquation;
  P: TPoint;
  X: Double;
  Y: Double;

  procedure DrawAxes;
  begin
    with Canvas do
    begin
      Pen.Color := FAxisColor;
      MoveTo(0, Height div 2);
      LineTo(Width, Height div 2);
      MoveTo(Width div 2, 0);
      LineTo(Width div 2, Height);
      Font.Assign(Font);
      TextOut(Width - TextWidth('X') - 2, Height div 2 + 6, 'X');
      TextOut(Width div 2 - TextWidth('Y') - 6, 2, 'Y');
    end;
  end;

  procedure DrawScale;
  begin
    with Canvas do
    begin
      Pen.Color := FAxisColor;
      X := -FXRange / 2;
      P.Y := Height div 2 - 5;
      while X < (FXRange / 2) do
      begin
        X := X + 1;
        P.X := Round(X / FScale) + Width div 2;
        MoveTo(P.X, P.Y);
        LineTo(P.X, P.Y + 10);
      end;
      Y := Round((Height div 2) * FScale) + 1;
      P.X := Width div 2 - 5;
      while P.Y < Height do
      begin
        Y := Y - 1;
        P.Y := -Round(Y / FScale) + Height div 2;
        MoveTo(P.X, P.Y);
        LineTo(P.X + 10, P.Y);
      end;
    end;
  end;

  procedure DrawGraph(Func: TFunc);
  begin
    P.X := 0;
    X := (P.X - Width div 2) * FScale;
    Y := Func(X);
    P.Y := -Round(Y / FScale) + Height div 2;
    Canvas.MoveTo(P.X, P.Y);
    while P.X < Width do
    begin
      Inc(P.X);
      X := (P.X - Width div 2) * FScale;
      Y := Func(X);
      P.Y := -Round(Y / FScale) + Height div 2;
      Canvas.LineTo(P.X, P.Y);
    end;
  end;

  procedure DrawZeroPoints;
  var
    Z: TDoubleArray;
    I: Integer;
  begin
    X := (Width div 2) * FScale;
    Equation.GetZeroPoints(-X, X, Z);
    for I := 0 to Length(Z) - 1 do
    begin
      P.X := Round(Z[I] / FScale) + Width div 2;
      P.Y := Height div 2;
      Canvas.Ellipse(P.X - 3, P.Y - 2, P.X + 4, P.Y + 4);
    end;
  end;

begin
  if (Width > 1) and (Height > 1) then
  begin
    Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect(0, 0, Width, Height));
    Canvas.Brush.Style := bsClear;
    if poAxes in FPaintOptions then
      DrawAxes;
    if poScale in FPaintOptions then
      DrawScale;
    for I := 0 to EquationCount - 1 do
    begin
      Equation := Equations[I];
      if Equation.Visible then
      begin
        Canvas.Pen.Color := Equation.Color;
        if poFormula in FPaintOptions then
          DrawGraph(Equation.Y);
        if poDerivative1 in FPaintOptions then
          DrawGraph(Equation.Y1);
        if poZeroPoints in FPaintOptions then
          DrawZeroPoints;
      end;
    end;
  end;
end;

procedure TGraph.Resize;
begin
  FScale := FXRange / Width;
  Invalidate;
  inherited Resize;
end;

procedure TGraph.SetAxisColor(Value: TColor);
begin
  if FAxisColor <> Value then
  begin
    FAxisColor := Value;
    Invalidate;
  end;
end;

procedure TGraph.SetPaintOptions(Value: TPaintOptions);
begin
  if FPaintOptions <> Value then
  begin
    FPaintOptions := Value;
    Invalidate;
  end;
end;

procedure TGraph.SetScale(Value: Single);
begin
  if FScale <> Value then
  begin
    FScale := Value;
    FXRange := FScale * Width;
    Invalidate;
  end;
end;

procedure TGraph.SetXRange(Value: Double);
begin
  if FXRange <> Value then
  begin
    FXRange := Value;
    FScale := FXRange / Width;
    Invalidate;
  end;
end;

end.
